import jinja2
import unidecode
import random
import uuid
import time

custom_filters = dict(
    ascii=lambda value: unidecode.unidecode(str(value)),
    shuffle=lambda value: ''.join(random.sample(str(value), len(str(value)))),
    chance=lambda value, chance: value if random.uniform(0.0, 1.0) <= chance else '',
    rjust=lambda value, width, fillchar = ' ': str(value).rjust(width, fillchar),
    ljust=lambda value, width, fillchar = ' ': str(value).ljust(width, fillchar),
    center=lambda value, width, fillchar=' ': str(value).center(width, fillchar),
)

custom_globals=dict(
    translate=lambda value, **dictionary: dictionary[value],
    uuid4=lambda: str(uuid.uuid4()),
    unixtime=time.time
)


def environment() -> jinja2.Environment:
    env = jinja2.Environment()

    env.filters.update(custom_filters)
    env.globals.update(custom_globals)

    return env



from unittest import TestCase
from unittest.mock import MagicMock
from fakr.TemplatedMapping import TemplatedMapping, templated_mapping
import collections


class TestTemplatedMapping(TestCase):

    renderer=MagicMock()
    mapping=dict(key1='value1', key2='value2', key3='value3')

    def setUp(self):
        self.subject=TemplatedMapping(self.renderer, **self.mapping)

    def test_is_mapping(self):
        self.assertIsInstance(self.subject, collections.Mapping)

    def test_len(self):
        self.assertIs(len(self.subject), len(self.mapping))

    def test_getitem(self):
        self.renderer.side_effect=lambda value, data: value

        for k, v in self.mapping.items():
            actual=self.subject[k]

            call_args=self.renderer.call_args[0]
            expected=(v, self.mapping)

            self.assertIs(actual, self.mapping[k])
            self.assertEqual(expected, call_args)

    def test_iter(self):
        self.assertIsInstance(iter(self.subject), collections.Iterator)

    def test_templated_mapping(self):
        mapping_factory=templated_mapping(self.renderer)
        self.assertTrue(callable(mapping_factory))

        subject=mapping_factory(**self.mapping)
        self.assertIsInstance(subject, TemplatedMapping)

        self.assertEqual(self.mapping, subject._TemplatedMapping__data)
        self.assertIs(self.renderer, subject._TemplatedMapping__render)


import collections
import time
import random


class Generator:
    def __init__(self, renderer: callable, vocabulary: collections.Sequence, template_string: str):
        self.__vocabulary=vocabulary
        self.__render=renderer
        self.__template_string=template_string

    def __call__(self, count: int, delay: float) -> collections.Iterator:
        counter=0
        while count==-1 or counter < count:
            time.sleep(delay)
            yield self.__render(self.__template_string, random.choice(self.__vocabulary), row=counter)
            counter += 1



import gzip
import collections
import json
import os

__vocabulary_search_paths=[
    os.path.dirname(os.path.realpath(__file__)) + '/vocabularies/{}',
    '{}',
    os.getcwd() + '/{}',
    os.getenv('HOME') + '/.fakr/{}',
]

def write(fp, data: collections.Sequence) -> None:
    jsoned=json.dumps(data).encode()
    gzipped=gzip.compress(bytes(jsoned))

    fp.write(gzipped)


def read(fp) -> collections.Sequence:
    gzipped=fp.read()
    jsoned=gzip.decompress(gzipped).decode()

    return json.loads(jsoned)


def search(file: str) -> str:
    for p in __vocabulary_search_paths:
        guess=os.path.realpath(p.format(file))
        if os.path.isfile(guess):
            return guess

    return file



import jinja2
from collections import Mapping
from functools import lru_cache


class Jinja2Renderer:
    __max_recursions=10

    def __init__(self, jinja2_env: jinja2.Environment, template_prefix: str=None) -> None:
        self.__jinja2_env=jinja2_env
        self.__tpl_prefix=template_prefix

    @lru_cache(maxsize=128)
    def __get_template_object(self, tpl_str: str) -> jinja2.Template:
        return self.__jinja2_env.from_string(tpl_str)

    def __render(self, value: str, mapping: Mapping, extra: Mapping, recursions=0) -> str:
        data=dict(extra)
        data.update(mapping)

        assert recursions <= self.__max_recursions # todo: Raise specific exception

        if self.__tpl_prefix is None:
            return self.__get_template_object(value).render(**data)

        if str(value).startswith(self.__tpl_prefix):
            return self.__render(
                self.__get_template_object(value[len(self.__tpl_prefix):]).render(**data),
                mapping,
                extra,
                recursions+1
            )

        return value

    def __call__(self, value: str, mapping: Mapping, **extra) -> str:
        return self.__render(value, mapping, extra, recursions=0)




import unittest
from unittest import mock
from fakr.Generator import Generator
import collections


class TestGenerator(unittest.TestCase):
    def setUp(self):
        self.template_string='template string'
        self.vocabulary=['vo', 'ca', 'bu', 'lary']
        self.renderer=mock.MagicMock()

        self.subject=Generator(self.renderer, self.vocabulary, self.template_string)

    def test_call_returns_iterator(self):
        self.assertIsInstance(self.subject(1, 1), collections.Iterator)

    @mock.patch('random.choice')
    @mock.patch('time.sleep')
    def test_iterator(self, mock_sleep, mock_choice):
        iterations=4
        delay=2

        self.renderer.side_effect = lambda t, v, **x: 'rendered'
        mock_choice.return_value='choice'
        subject=self.subject(iterations, delay)

        for row, item in enumerate(subject):
            self.renderer.assert_called_with(self.template_string, 'choice', row=row)
            mock_sleep.assert_called_with(delay)
            self.assertEqual('rendered', item)

        self.assertEqual(iterations, self.renderer.call_count)
        self.assertEqual(iterations, mock_sleep.call_count)




from collections import Mapping, Sequence
import argparse
import sys
import os
from .Jinja2Renderer import Jinja2Renderer
from .CompoundMappingSequence import CompoundMappingSequence
from .TemplatedMapping import templated_mapping
from .jinja import environment
import jinja2.exceptions
from . import version, package_name, default_vocabulary, vocabulary
from .Generator import Generator

ENV_FAKR_VOCABUALRY= 'FAKR_VOCABULARY'
ENV_FAKR_MIXIN= 'FAKR_MIXIN'


def main():
    j2env = environment()
    args = __parse_args()

    vocabulary_fps=args['vocabulary'] + args['mixin']
    vocabulary = load_vocabulary(j2env, *vocabulary_fps)

    if args['list'] is True:
        sys.stdout.write('\n'.join(sorted([v for v in list(vocabulary[0].keys()) + ['row']])) + '\n')
        exit(0)

    if args['info'] is True:
        try:
            voclen=len(vocabulary)
        except OverflowError:
            voclen='Too many'

        sys.stdout.write('{} unique entries in\n{}\n'.format(voclen, '\n'.join(['\t- ' + os.path.realpath(fp.name) for fp in vocabulary_fps])))
        exit(0)

    try:
        generator = Generator(Jinja2Renderer(j2env, template_prefix=None), vocabulary, sys.stdin.read())
        for item in generator(args['count'], args['delay']):
            sys.stdout.write(item + '\n')
    except KeyboardInterrupt:
        exit(0)
    except jinja2.exceptions.TemplateError as e:
        sys.stderr.write('There was a problem with your templates: {}'.format(e.message))
    except OSError as e:
        sys.stderr.write(e.strerror)


def load_vocabulary(j2env, *fps) -> Sequence:
    partitions=list()
    for fp in fps:
        partitions+=vocabulary.read(fp)

    mapping_factory=templated_mapping(Jinja2Renderer(j2env, template_prefix='%%'))
    return CompoundMappingSequence(mapping_factory, *partitions)


def __parse_args() -> Mapping:
    parser = argparse.ArgumentParser(
        prog=package_name,
        description='''
            {} reads a jinja2 template from STDIN or from a file, renders it with random values from a vocabulary or from a custom file and writes the result to STDOUT
        '''.format(package_name)
    )

    meta_group=parser.add_argument_group('Information')

    meta_group.add_argument('--version',
                        action='version',
                        version='%(prog)s {}'.format(version)
                        )

    parser.add_argument('-c', '--count',
                        metavar='COUNT',
                        type=int,
                        default=-1,
                        help='Number of datasets to generate, defaults to 0 (unlimited)'
                        )

    parser.add_argument('-d', '--delay',
                        metavar='SECONDS',
                        type=float,
                        default=0,
                        help='Wait SECONDS between dataset generation, defaults to 0 (no waiting)'
                        )

    meta_group.add_argument('-l', '--list',
                        action='store_true',
                        help='show the available variables and exit'
                        )

    meta_group.add_argument('-i', '--info',
                        action='store_true',
                        help='show some vocabulary information and exit'
                        )

    parser.add_argument('-v', '--vocabulary',
                        metavar='FILENAME',
                        action=VocabularyFileAction,
                        help='Path to the vocabulary file. Defaults to the builtin vocabulary "{}". This setting overrides the vocabulary selection via the environment variable FAKR_VOCABULARY'.format(os.path.basename(default_vocabulary)),
                        default=VocabularyFileAction.openFiles(os.getenv(ENV_FAKR_VOCABUALRY, default_vocabulary)))

    parser.add_argument('-m', '--mixin',
                        metavar='FILENAME',
                        action=VocabularyFileAction,
                        help='Additional vocabulary files to mix into the main vocabulary. This setting overrides the vocabulary selection via the environment variable FAKR_MIXIN',
                        nargs='*',
                        default=VocabularyFileAction.openFiles(*[v.strip() for v in filter(None, os.getenv(ENV_FAKR_MIXIN, '').split(':'))])
                        )

    return vars(parser.parse_args())


class VocabularyFileAction(argparse.Action):

    @staticmethod
    def openFiles(*files: str, resolver: callable = lambda v: v) -> list:
        fps=[]
        for file in files:
            try:
                fp = open(resolver(file), 'rb')
                fps.append(fp)
            except FileNotFoundError as e:
                print('{}: {}'.format(file, e.strerror))
                exit(1)

        return fps

    def __call__(self, parser, args, values, option_string=None):
        if type(values) is str:
            values=[values]

        setattr(args, self.dest, self.openFiles(*values, resolver=vocabulary.search))


from .__version__ import version
import os

package_name='fakr'
repo_url='http://github.com/l-x/{}'.format(package_name)
download_url="{}/archive/{}.tar.gz".format(repo_url, version)
author = 'l-x'
author_email = 'l-x@mailbox.org'

default_vocabulary=os.path.dirname(os.path.realpath(__file__)) + '/vocabularies/us_business.fakr'



import csv
import argparse
from fakr import version, package_name as main_package_name
from collections import Mapping

package_name=main_package_name + '-builder'


def main():
    args=dict(
        delimiter=',',
        quotechar='"'
    )
    args.update(__parse_args())
    vocabulary_data=list()
    for fp in args['files']:
        partition=list()
        reader=csv.DictReader(fp, delimiter=args['delimiter'], quotechar=args['quotechar'])
        for l in reader:
            partition.append(l)
        vocabulary_data.append(partition)

    from . import vocabulary

    vocabulary.write(args['output'], vocabulary_data)
    args['output'].close()

def __parse_args() -> Mapping:
    parser = argparse.ArgumentParser(
        prog=package_name,
        description='''
            {} takes a list of csv files and creates a vocabulary
        '''.format(package_name)
    )

    parser.add_argument('--version',
                        action='version',
                        version='%(prog)s {}'.format(version)
                        )

    parser.add_argument(metavar='FILENAME',
                        dest='output',
                        type=argparse.FileType('wb'),
                        help='Filename for output',
                        )

    parser.add_argument('-t', '--tables',
                        metavar='FILENAME',
                        dest='files',
                        type=argparse.FileType('r'),
                        help='Name of a csv file',
                        nargs='+',
                        default=[]
                        )

    return vars(parser.parse_args())



import unittest
from unittest import mock
from fakr import jinja


class TestCustomFilters(unittest.TestCase):

    def setUp(self):
        self.subject=jinja.custom_filters

    @mock.patch('unidecode.unidecode', side_effect=lambda v: v)
    def test_ascii(self, mock_unidecode: mock.MagicMock):
        test_value='value'
        value=mock.MagicMock()
        value.__str__.return_value=test_value

        self.assertIs(test_value, self.subject['ascii'](value))
        value.__str__.assert_called_once_with()
        mock_unidecode.assert_called_once_with(test_value)

    @mock.patch('random.sample', side_effect=lambda v, _: v)
    def test_shuffle(self, mock_sample: mock.MagicMock):
        test_value='value'
        value=mock.MagicMock()
        value.__str__.return_value=test_value

        self.assertIn(test_value, self.subject['shuffle'](value))
        value.__str__.assert_called_with()
        mock_sample.assert_called_once_with(test_value, len(test_value))

    def test_chance(self):
        def assertChance(test_value: str, test_chance: float, expected_value: str):
            random_uniform_result=0.5
            with mock.patch('random.uniform', side_effect=lambda f, t: random_uniform_result) as mock_uniform:
                self.assertIs(expected_value, self.subject['chance'](test_value, test_chance))
                mock_uniform.assert_called_once_with(0.0, 1.0)

        assertChance('value', 0.0, '')
        assertChance('value', 0.49999999999, '')
        assertChance('value', 0.5, 'value')
        assertChance('value', 1.0, 'value')
        assertChance('value', 1234, 'value')

    def assertStringFunction(self, filter: str, value: str, width: int, fillchar: str, expected: str):
        value_mock=mock.MagicMock()
        value_mock.__str__.return_value=value

        self.assertEqual(expected, self.subject[filter](value_mock, width, fillchar))
        value_mock.__str__.assert_called_once_with()

    def test_ljust(self):
        self.assertStringFunction('ljust', 'v', 3, 'x', 'vxx')

    def test_rjust(self):
        self.assertStringFunction('rjust', 'v', 3, 'x', 'xxv')

    def center(self):
        self.assertStringFunction('center', 'v', 3, 'x', 'xvx')


class TestCustomGlobals(unittest.TestCase):

    def setUp(self):
        self.subject=jinja.custom_globals

    def test_translate(self):
        test_value='test_value'
        expected_value='expected_value'

        self.assertEqual(expected_value, self.subject['translate'](test_value, test_value=expected_value))

    def test_unixtime(self):
        import time
        self.assertIs(self.subject['unixtime'], time.time)

    @mock.patch('uuid.uuid4')
    def test_uuid4(self, mock_uuid: mock.MagicMock):
        uuid_value='test-uuid'
        uuid4=mock.MagicMock()
        uuid4.__str__.return_value=uuid_value
        mock_uuid.return_value=uuid4

        self.assertEqual(uuid_value, self.subject['uuid4']())
        uuid4.__str__.assert_called_once_with()
        mock_uuid.assert_called_once_with()


class TestEnvironment(unittest.TestCase):

    @mock.patch('jinja2.Environment')
    def test_environment(self, mock_env):
        env=mock_env.return_value

        self.assertIs(env, jinja.environment())
        env.filters.update.assert_called_once_with(jinja.custom_filters)
        env.globals.update.assert_called_once_with(jinja.custom_globals)


from setuptools import setup
import io
import sys
from fakr import version, repo_url, download_url, package_name, author, author_email

with io.open('README.rst', 'r', encoding='utf-8') as readme_file:
    readme = readme_file.read()

if sys.argv[-1] == 'readme':
    print(readme)
    sys.exit()

requirements = [
    'unidecode',
    'jinja2'
]

setup(
    name=package_name,
    version=version,
    description='Random data generator for the command line',
    long_description=readme,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3 :: Only',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Topic :: Software Development',
        'Topic :: Software Development :: Testing',
        'Topic :: Software Development :: Testing :: Traffic Generation',
    ],
    keywords='fake template random list data generator Jinja2 ' + package_name,
    url=repo_url,
    author=author,
    author_email=author_email,
    license='MIT',
    install_requires=open('requirements.txt').read().lstrip().rstrip().split('\n'),
    tests_require=['nose2', 'cov-core'],
    test_suite='nose2.collector.collector',
    entry_points={
        'console_scripts': [
            '{n}={n}.cli:main'.format(n=package_name),
            '{n}-builder={n}.vocabulary_builder:main'.format(n=package_name),
        ],
    },
    include_package_data=True,
    zip_safe=False,
    packages=[package_name],
    package_dir={package_name: package_name + '/'},
    package_data={package_name: ['vocabularies/*.fakr']},
    download_url=download_url
)



import unittest
from fakr.CompoundMappingSequence import CompoundMappingSequence


class TestCompoundMappingSequence(unittest.TestCase):
    partitions=[
        [
            dict(a='1', b='2', c='3'),
            dict(a='4', b='5', c='6'),
        ],
        [
            dict(d='1', e='2', f='3'),
            dict(d='4', e='5', f='6'),
        ],
    ]

    expected=[
        dict(id=0, guid='cfcd2084-95d5-65ef-66e7-dff9f98764da', a='1', b='2', c='3', d='1', e='2', f='3'),
        dict(id=1, guid='c4ca4238-a0b9-2382-0dcc-509a6f75849b', a='4', b='5', c='6', d='1', e='2', f='3'),
        dict(id=2, guid='c81e728d-9d4c-2f63-6f06-7f89cc14862c', a='1', b='2', c='3', d='4', e='5', f='6'),
        dict(id=3, guid='eccbc87e-4b5c-e2fe-2830-8fd9f2a7baf3', a='4', b='5', c='6', d='4', e='5', f='6'),
    ]

    sequence_length=4

    def setUp(self):
        self.mapping_factory=dict

        self.subject=CompoundMappingSequence(self.mapping_factory, *self.partitions)

    def test_len(self):
        self.assertIs(self.sequence_length, len(self.subject))

    def test_getitem_raises_index_error(self):
        try:
            self.subject[self.sequence_length+1]
        except IndexError as e:
            return

        self.fail('Expected IndexError was not raised')

    def test_getitem(self):
        for k, v in enumerate(self.subject):
            self.assertEqual(v, self.expected[k])